# Copyright 2015-2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import os
import re
import select
import socket
import testtools
import time

from itertools import chain
from networking_fujitsu.ml2.cfab import cfabdriver
from networking_fujitsu.ml2.cfab.mech_cfab import cfg
from neutron.common import utils
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import config as ml2_config
from neutron.tests import base

FUJITSU_CFAB = "networking_fujitsu.ml2.cfab."
_CFABDRIVER__CFABMANAGER = FUJITSU_CFAB + "cfabdriver._CFABManager"
_TELNETLIB_TELNET = FUJITSU_CFAB + "cfabdriver.telnetlib.Telnet"
_EXCLUDE_BRACKET_LINE_RE = re.compile(r"^[^[].*$", re.MULTILINE)


class BaseTestMockedCFABManager(base.BaseTestCase):
    """Base class to test Fujitsu C-Fabric manager."""

    def setUp(self):
        super(BaseTestMockedCFABManager, self).setUp()
        self.manager = cfabdriver._CFABManager()
        self.manager.close_session = mock.MagicMock()

    def assert_wrote(self, lines):
        telnet = self.manager._telnet
        """:type : mock.MagicMock"""
        self.assertEqual(
            lines, [x[0][0] for x in telnet.write.call_args_list])


class TestMockedCFABManager(BaseTestMockedCFABManager):
    """Test Fujitsu C-Fabric manager."""

    def test_connect(self):
        with mock.patch(_TELNETLIB_TELNET, autospec=True) as telnet:

            self.manager.connect("address", "username", "password")

        telnet.assert_called_once_with(
            host="address",
            port=cfabdriver.TELNET_PORT,
            timeout=cfabdriver._TIMEOUT)
        self.assert_wrote(["username\n", "password\n"])

    def test_connect_fail(self):
        with mock.patch(_TELNETLIB_TELNET, autospec=True) as telnet:
            telnet.side_effect = socket.error

            self.assertRaises(
                socket.error,
                self.manager.connect, "address", "username", "password")

    def test_reconnect_raise_exceptions(self):

        with mock.patch(_TELNETLIB_TELNET, autospec=True) as telnet:
            for er in [EOFError, EnvironmentError, ValueError, OSError]:
                tel = telnet.return_value
                tel.read_until.side_effect = er
                self.manager.close_session()
                self.assertRaises(er, self.manager._reconnect)
                self.assertEqual(0, self.manager._retry_count)

    def test_reconnect_busy_and_retry(self):

        busy = 'The system is busy. Please login after waiting for a while.\n'
        max_session = 'Login failed to switch.(too many sessions. bye!\n)'
        with mock.patch(_TELNETLIB_TELNET, autospec=True) as telnet:
            tel = telnet.return_value
            tel.read_until.side_effect = [busy,
                                          max_session,
                                          cfabdriver._PROMPT_LOGIN,
                                          cfabdriver._PROMPT_PASS,
                                          cfabdriver._PROMPT_ADMIN]
            time.sleep = mock.MagicMock()
            time.sleep.side_effect = None

            self.manager.connect("address", "username", "password")
            self.assertEqual(3, self.manager.close_session.call_count)
            self.assertEqual(0, self.manager._retry_count)
            time.sleep.assert_called_with(cfabdriver._WAIT_FOR_BUSY)

    def test_reconnect_busy_and_reached_maxium_retry(self):

        busy = 'The system is busy. Please login after waiting for a while.\n'
        with mock.patch(_TELNETLIB_TELNET, autospec=True) as telnet:
            tel = telnet.return_value
            tel.read_until.return_value = busy
            time.sleep = mock.MagicMock()
            time.sleep.side_effect = None

            self.assertRaises(
                ValueError,
                self.manager.connect, "address", "username", "password")
            retry_count = cfabdriver._TIMEOUT / cfabdriver._WAIT_FOR_BUSY
            self.assertEqual(12, self.manager.close_session.call_count)
            self.assertEqual(retry_count, time.sleep.call_count)
            self.assertEqual(0, self.manager._retry_count)


class BaseTestMockedCFABManagerConnected(BaseTestMockedCFABManager):
    """Base class to test Fujitsu C-Fabric manager after connected."""

    def setUp(self):
        super(BaseTestMockedCFABManagerConnected, self).setUp()
        with mock.patch(_TELNETLIB_TELNET, autospec=True):
            self.manager.connect("address", "username", "password")
        self.prompt = "# "

        def read_until(*args, **kwargs):
            return "(config)# "

        def expect(*args, **kwargs):
            s = self.prompt
            m = args[0][0].search(s)
            return 0 if m is not None else -1, m, s

        self.manager._telnet.read_until.side_effect = read_until
        self.manager._telnet.expect.side_effect = expect


class TestMockedCFABManagerConnected(BaseTestMockedCFABManagerConnected):
    """Test Fujitsu C-Fabric manager after connected.

    """

    def test_get_candidate_config(self):
        candidate_config = self.manager.get_candidate_config()

        self.assertEqual(
            mock.call("show candidate-config\n"),
            self.manager._telnet.write.call_args)
        self.assertEqual("", candidate_config)

    def test_get_running_config(self):
        running_config = self.manager.get_running_config()

        self.assertEqual(
            mock.call("show running-config\n"),
            self.manager._telnet.write.call_args)
        self.assertEqual("", running_config)

    def test_configure(self):
        cmd = "pprofile 1 vlan tag 1"

        self.manager.configure([cmd])

        call_args_list = self.manager._telnet.write.call_args_list
        self.assertIn(mock.call("configure\n"), call_args_list)
        self.assertIn(mock.call(cmd + "\n"), call_args_list)
        self.assertIn(mock.call("commit\n"), call_args_list)
        self.assertEqual(
            mock.call("save\n"), self.manager._telnet.write.call_args)

    def test_configure_without_commit(self):
        cmd = "pprofile 1 vlan tag 1"

        self.manager.configure([cmd], commit=False)

        call_args_list = self.manager._telnet.write.call_args_list
        self.assertIn(mock.call("configure\n"), call_args_list)
        self.assertIn(mock.call(cmd + "\n"), call_args_list)
        self.assertNotIn(mock.call("commit\n"), call_args_list)


class TestMockedCFABManagerConnectedWithoutSave(
        BaseTestMockedCFABManagerConnected):
    """Test Fujitsu C-Fabric manager after connected without save.

    """

    def setUp(self):
        super(TestMockedCFABManagerConnectedWithoutSave, self).setUp()
        self.manager.save_config = False

    def test_configure_without_save(self):
        cmd = "pprofile 1 vlan tag 1"

        self.manager.configure([cmd])

        call_args_list = self.manager._telnet.write.call_args_list
        self.assertIn(mock.call("configure\n"), call_args_list)
        self.assertIn(mock.call(cmd + "\n"), call_args_list)
        self.assertEqual(
            mock.call("commit\n"), self.manager._telnet.write.call_args)
        self.assertNotIn(mock.call("save\n"), call_args_list)


@testtools.skipUnless(
    'OS_FUJITSU_CFAB_ADDRESS' in os.environ,
    "OS_FUJITSU_CFAB_ADDRESS environment variable is not defined.")
class TestCFABManager(base.BaseTestCase):
    """Test Fujitsu C-Fabric manager using the real telnet connection.

    Tests will be performed using the C-Fabric CLI through a telnet connection
    to the address OS_FUJITSU_CFAB_ADDRESS and the port OS_FUJITSU_CFAB_PORT
    (defaults to 23). The username will be taken from OS_FUJITSU_CFAB_USERNAME
    (defaults to "admin") and the password will be taken from
    OS_FUJITSU_CFAB_PASSWORD (defaults to "password").
    If the environment variable OS_FUJITSU_CFAB_ADDRESS is NOT defined, tests
    will be skipped.

    """

    def _setup_lock(self):
        """Set up lock_path so that all tests are serialized.

        This is necessary to keep the C-Fabric config consistent within each
        test.
        """

        try:
            ml2_config.cfg.CONF.set_override('lock_path', "lock")
        except ml2_config.cfg.NoSuchOptError:
            ml2_config.cfg.CONF.set_override(
                'lock_path', "lock", "oslo_concurrency")

    def setUp(self):
        super(TestCFABManager, self).setUp()
        self._setup_lock()
        try:
            cfabdriver.TELNET_PORT = int(os.environ['OS_FUJITSU_CFAB_PORT'])
        except KeyError:
            pass
        self.manager = cfabdriver.CFAB_MANAGER
        self.manager.connect(
            os.environ.get('OS_FUJITSU_CFAB_ADDRESS'),
            os.environ.get('OS_FUJITSU_CFAB_USERNAME') or "admin",
            os.environ.get('OS_FUJITSU_CFAB_PASSWORD') or "password",
        )

    def assert_running_config(self, prefix, expected_config):
        running_config = self.manager.get_running_config(prefix=prefix)
        self.assertEqual(
            expected_config, _EXCLUDE_BRACKET_LINE_RE.findall(running_config))

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_modes(self):
        self.manager._close_session()
        self.assertEqual(cfabdriver._MODE_ADMIN, self.manager._get_mode())
        self.manager._execute("configure")
        self.assertEqual(cfabdriver._MODE_CONFIG, self.manager._get_mode())
        self.manager._execute("interface 1/1/1/1")
        self.assertEqual(cfabdriver._MODE_CONFIG_IF, self.manager._get_mode())
        self.manager._execute("exit")
        self.assertEqual(cfabdriver._MODE_CONFIG, self.manager._get_mode())
        self.manager._execute("exit")
        self.assertEqual(cfabdriver._MODE_ADMIN, self.manager._get_mode())
        self.manager._execute("exit")
        self.assertEqual(cfabdriver._MODE_USER, self.manager._get_mode())

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_get_running_config(self):
        self.manager.configure(
            ["no pprofile",
             "pprofile 1 vlan tag 1",
             "pprofile 2 vlan tag 2"])

        running_config = self.manager.get_running_config()

        self.assertEqual(
            ["pprofile 1 vlan tag 1", "pprofile 2 vlan tag 2"],
            re.findall(r"^pprofile\s+.+$", running_config, re.MULTILINE))

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_get_running_config_prefix(self):
        self.manager.configure(
            ["no pprofile",
             "pprofile 1 vlan tag 1",
             "pprofile 2 vlan tag 2"])

        self.assert_running_config(
            "pprofile", ["1 vlan tag 1", "2 vlan tag 2"])

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_configure(self):
        self.manager.configure(["no pprofile"])

        self.assert_running_config("pprofile", [])

        self.manager.configure(["pprofile 1 vlan tag 1"])

        self.assert_running_config("pprofile", ["1 vlan tag 1"])

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_configure_from_interface_config(self):
        self.manager.configure(["no pprofile"])
        self.manager._execute("interface 1/1/1/1")

        self.manager.configure(["pprofile 1 vlan tag 1"])

        self.assert_running_config("pprofile", ["1 vlan tag 1"])

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_configure_from_user(self):
        self.manager.configure(["no pprofile"])
        self.manager._execute("exit")
        self.manager._execute("exit")

        self.manager.configure(["pprofile 1 vlan tag 1"])

        self.assert_running_config("pprofile", ["1 vlan tag 1"])

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_configure_from_closed(self):
        self.manager.configure(["no pprofile"])
        self.manager._close_session()

        self.manager.configure(["pprofile 1 vlan tag 1"])

        self.assert_running_config("pprofile", ["1 vlan tag 1"])

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_configure_no_commit(self):
        self.manager.configure(["no pprofile"])

        self.manager.configure(["pprofile 1 vlan tag 1"], commit=False)

        self.assert_running_config("pprofile", [])

        self.manager.configure([])

        self.assert_running_config("pprofile", ["1 vlan tag 1"])

    @utils.synchronized(cfabdriver._LOCK_NAME, external=True)
    def test_configure_error(self):
        self.assertRaises(
            ml2_exc.MechanismDriverError, self.manager.configure, ["error"])


class BaseTestCFABdriver(base.BaseTestCase):

    """Base class to test Fujitsu C-Fabric mechanism driver.

    """

    def setUp(self):
        self.ports = "1/1/0/1"
        self.mac = "00:01:02:03:04:05"
        super(BaseTestCFABdriver, self).setUp()
        with mock.patch(_CFABDRIVER__CFABMANAGER, autospec=True) as mocked:
            self.driver = cfabdriver.CFABdriver(cfg.CONF)
            self.driver.mgr = mocked.return_value

    def assert_configured(self, cmds):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        tmp = [x[0][0] for x in mgr.configure.call_args_list]
        actual = list(chain.from_iterable(tmp))
        self.assertEqual(cmds, actual)

        # Make sure that only the last configure has commit=True.
        commits = [x[1].get('commit', True)
                   for x in mgr.configure.call_args_list]
        self.assertTrue(commits.pop())
        commits.append(False)
        self.assertEqual({False}, set(commits))


class TestCFABdriver(BaseTestCFABdriver):

    """Test Fujitsu C-Fabric mechanism driver.

    """

    def test_associate_mac_to_network_raises(self):
        self.driver.mgr = mock.Mock()
        mgr = self.driver.mgr
        cfab = self.driver
        for er in [EOFError, EnvironmentError, OSError, select.error]:
            mgr.connect.side_effect = er
            self.assertRaises(er, cfab.associate_mac_to_network,
                              'a', 'u', 'p', '1', 8, self.mac)
        mgr.connect.side_effect = ml2_exc.MechanismDriverError(
            method='connect')
        self.assertRaises(ml2_exc.MechanismDriverError,
                          cfab.associate_mac_to_network,
                          'a', 'u', 'p', '1', 8, self.mac)
        self.assertEqual(5, mgr.close_session.call_count)

    def test_disassociate_mac_from_network_raises(self):
        self.driver.mgr = mock.Mock()
        mgr = self.driver.mgr
        cfab = self.driver
        for er in [EOFError, EnvironmentError, OSError, select.error]:
            mgr.connect.side_effect = er
            self.assertRaises(er, cfab.dissociate_mac_from_network,
                              'a', 'u', 'p', '1', 8, self.mac)
        mgr.connect.side_effect = ml2_exc.MechanismDriverError(
            method='connect')
        self.assertRaises(ml2_exc.MechanismDriverError,
                          cfab.dissociate_mac_from_network,
                          'a', 'u', 'p', '1', 8, self.mac)
        self.assertEqual(5, mgr.close_session.call_count)

    def test_associate_mac_to_network(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = (
            """pprofile 00:01:02:03:04:05 vlan tag 2
vfab 4 pprofile 0 vsiid mac 00:01:02:03:04:05 00:01:02:03:04:05
""")

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        mgr.close_session.assert_called_once_with()
        self.assert_configured(
            ["vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 "
             "00:01:02:03:04:05"])

    def test_associate_mac_to_network_no_pprofile(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = ""

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["pprofile 00:01:02:03:04:05 vlan tag 2",
             "vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 "
             "00:01:02:03:04:05"])

    def test_associate_mac_to_network_existing(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = (
            """pprofile 00:01:02:03:04:05 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 00:01:02:03:04:05
""")

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assertFalse(mgr.configure.called)

    def test_associate_mac_to_network_existing_override(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = (
            """pprofile test-2 vlan tag 2
vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:05 test-2
""")

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["pprofile 00:01:02:03:04:05 vlan tag 2",
             "vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:05 "
             "00:01:02:03:04:05"])

    def test_associate_mac_to_network_override_pprofile(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = (
            """pprofile 00:01:02:03:04:05 vlan tag 1,2
vfab 4 pprofile 0 vsiid mac 00:01:02:03:04:05 00:01:02:03:04:05
""")

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["pprofile 00:01:02:03:04:05 vlan tag 2",
             "vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 "
             "00:01:02:03:04:05"])

    def test_dissociate_mac_from_network(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = (
            """pprofile 00:01:02:03:04:05 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 00:01:02:03:04:05
""")

        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["no vfab 3 pprofile 0",
             "no pprofile 00:01:02:03:04:05"])

    def test_dissociate_mac_from_network_still_used_in_other_vfab(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = (
            """pprofile 00:01:02:03:04:05 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 00:01:02:03:04:05
vfab 4 pprofile 0 vsiid mac 00:01:02:03:04:05 00:01:02:03:04:05
""")

        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(["no vfab 3 pprofile 0"])


class TestCFABdriverSharePprofile(BaseTestCFABdriver):
    """Test Fujitsu C-Fabric mechanism driver with shared pprofile.

    """
    def setUp(self):
        cfg.CONF.set_override('share_pprofile', True, "fujitsu_cfab")
        super(TestCFABdriverSharePprofile, self).setUp()

    def test_associate_mac_to_network(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 2 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:00:00:00:00:01 2
"""

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:05 2"])

    def test_associate_mac_to_network_no_pprofile(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = ""

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        mgr.close_session.assert_called_once_with()
        self.assert_configured(
            ["pprofile 2 vlan tag 2",
             "vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 2"])

    def test_associate_mac_to_network_existing(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 1
"""

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assertFalse(mgr.configure.called)

    def test_associate_mac_to_network_existing_override(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 4
vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:05 1
"""

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["pprofile 2 vlan tag 2",
             "vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:05 2"])

    def test_dissociate_mac_from_network(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 1
"""

        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["no vfab 3 pprofile 0",
             "no pprofile 1"])

    def test_dissociate_mac_from_network_still_used(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 1
vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:06 1
"""

        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(["no vfab 3 pprofile 0"])

    def test_dissociate_mac_from_network_still_used_in_other_vfab(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 1
vfab 4 pprofile 0 vsiid mac 00:01:02:03:04:06 1
"""

        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(["no vfab 3 pprofile 0"])

    def test_dissociate_mac_from_network_no_match(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 4
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 1
"""

        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assertFalse(mgr.configure.called)


class TestCFABdriverSharedPprofilePrefixed(BaseTestCFABdriver):
    """Test Fujitsu C-Fabric mechanism driver with pprofile prefix.

    """
    def setUp(self):
        cfg.CONF.set_override('share_pprofile', True, "fujitsu_cfab")
        cfg.CONF.set_override('pprofile_prefix', "test-", "fujitsu_cfab")
        super(TestCFABdriverSharedPprofilePrefixed, self).setUp()

    def test_associate_mac_to_network(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile test-2 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:00:00:00:00:01 test-2
"""

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.close_session.assert_called_once_with()
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:05 test-2"])

    def test_associate_mac_to_network_no_pprofile(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = ""

        self.driver.associate_mac_to_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["pprofile test-2 vlan tag 2",
             "vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 test-2"])

    def test_dissociate_mac_from_network(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 2
pprofile test-1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 test-1
vfab 4 pprofile 0 vsiid mac 00:01:02:03:04:06 1
"""
        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(
            ["no vfab 3 pprofile 0",
             "no pprofile test-1"])

    def test_dissociate_mac_from_network_still_used(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile test-1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 test-1
vfab 3 pprofile 1 vsiid mac 00:01:02:03:04:06 test-1
"""
        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(["no vfab 3 pprofile 0"])

    def test_dissociate_mac_from_network_still_used_in_other_vfab(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile test-1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 test-1
vfab 4 pprofile 0 vsiid mac 00:01:02:03:04:06 test-1
"""
        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assert_configured(["no vfab 3 pprofile 0"])

    def test_dissociate_mac_from_network_no_match(self):
        mgr = self.driver.mgr
        """:type : mock.MagicMock"""
        mgr.get_running_config.return_value = """pprofile 1 vlan tag 2
vfab 3 pprofile 0 vsiid mac 00:01:02:03:04:05 1
"""
        self.driver.dissociate_mac_from_network(
            "address", "username", "password", "3", 2, "00:01:02:03:04:05")

        mgr.connect.assert_called_once_with("address", "username", "password")
        mgr.get_running_config.assert_called_once_with()
        self.assertFalse(mgr.configure.called)


class TestCFABdriverPprofilePrefix(base.BaseTestCase):
    """Test Fujitsu C-Fabric mechanism driver for pprofile_prefix errors.

    """

    def test_too_long(self):
        cfg.CONF.set_override('pprofile_prefix', "a" * 29, "fujitsu_cfab")
        with mock.patch(_CFABDRIVER__CFABMANAGER, autospec=True):
            self.assertRaises(ValueError, cfabdriver.CFABdriver, cfg.CONF)

    def test_illegal_character(self):
        cfg.CONF.set_override('pprofile_prefix', '"', "fujitsu_cfab")
        with mock.patch(_CFABDRIVER__CFABMANAGER, autospec=True):
            self.assertRaises(ValueError, cfabdriver.CFABdriver, cfg.CONF)


class TestCFABdriverSetupVlan(BaseTestCFABdriver):
    """Test Fujitsu C-Fabric mechanism driver for VLAN configuration.

    """

    def setUp(self):
        cfg.CONF.set_override('pprofile_prefix', "test-", "fujitsu_cfab")
        super(TestCFABdriverSetupVlan, self).setUp()

    def test_raises(self):
        mgr = self.driver.mgr
        cfab = self.driver
        for er in [EOFError, EnvironmentError, OSError, select.error]:
            mgr.get_candidate_config.side_effect = er
            self.assertRaises(er,
                              cfab.setup_vlan, 'a', 'u', 'p', '1', 8,
                              self.ports, self.mac)
        self.assertEqual(4, mgr.close_session.call_count)

    def test_no_preconfig_exist(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = "no_preconfig"
        expect = []
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_already_configured_vlan_but_missing_interface(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    exit
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'delete') \
            + cfab_cmd('interface', 'add', ports=self.ports) \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_already_configured_vlan(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'delete') \
            + cfab_cmd('interface', 'add', ports=self.ports) \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_already_configured_vlan_and_includes_boundary(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
ifgroup 1 ether 1/1/0/2
ifgroup 2 ether 1/1/0/3
interface 1/1/0/1
    exit
vfab 1 vlan 8 endpoint untag 0-2
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('vlan', 'replace', ifg='1-2') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1-2,0')
        self.assert_configured(expect)

    def test_already_configured_vlan_and_includes_between_boundary(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/2
ifgroup 1 ether 1/1/0/1
ifgroup 2 ether 1/1/0/3
interface 1/1/0/1
    exit
vfab 1 vlan 8 endpoint untag 0-2
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('vlan', 'replace', ifg='0,2') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='0,2,1')
        self.assert_configured(expect)

    def test_already_configured_vlan_and_not_included_in_boundary(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/2
ifgroup 1 ether 1/1/0/3
ifgroup 2 ether 1/1/0/4
interface 1/1/0/1
    exit
vfab 1 vlan 8 endpoint untag 0-2
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='3') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='0-2,3')
        self.assert_configured(expect)

    def test_exists_other_vlan_definition_missing_interface_def(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    exit
vfab 1 vlan 100 endpoint untag 0
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('vlan', 'delete', vlanid=100) \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exists_other_vlan_definition(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
vfab 1 vlan 100 endpoint untag 0
        """

        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('vlan', 'delete', vlanid=100) \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exists_other_vlan_definition_with_different_ifgroup(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
ifgroup 100 ether 1/1/0/1
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
vfab 1 vlan 100 endpoint untag 100
        """

        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('vlan', 'delete', vlanid=100) \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exists_interface_def_for_target_port(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.close_session.assert_called_once_with()
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exists_interface_def_for_other_port(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
interface 1/1/0/2
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exists_interface_def_of_lag(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
interface 1/1/0/1
    type linkaggregation 1
    exit
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('lag', 'delete') \
            + cfab_cmd('ifgroup', 'add') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exist_ifgroup(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = "ifgroup 0 ether 1/1/0/10"
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exist_ifgroup_and_reuse(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = "ifgroup 0 ether 1/1/0/1"
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exist_ifgroup_with_port_range(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1-1/1/0/5
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exist_ifgroup_with_port_range_and_out_of_range(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/2-1/1/0/5
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exist_ifgroup_with_lag(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = "ifgroup 0 linkaggregation 1 1"
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exist_ifgroup_with_lag_range(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1-4
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exists_lag_without_interface(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    exit
interface 1/1/0/2
    exit
linkaggregation 1 1 type endpoint
linkaggregation 1 1 mode active
linkaggregation 1 1 cfab port-mode external
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exist_definition_type_is_lag(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    exit
interface 1/1/0/2
    type linkaggregation 1
    exit
linkaggregation 1 1 type endpoint
linkaggregation 1 1 mode active
linkaggregation 1 1 cfab port-mode external
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('lag', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exist_lag_and_vlan_definition(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    exit
interface 1/1/0/2
    type linkaggregation 1
    exit
linkaggregation 1 1 type endpoint
linkaggregation 1 1 mode active
linkaggregation 1 1 cfab port-mode external
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.setup_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete') \
            + cfab_cmd('vlan', 'delete') \
            + cfab_cmd('lag', 'delete') \
            + cfab_cmd('ifgroup', 'add', ifg='1') \
            + cfab_cmd('interface', 'add') \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_ifgroup_ether_is_exhauted(self):
        mgr = self.driver.mgr
        candidate = ""
        for i in range(0, 4096):
            candidate += 'ifgroup {if_id} ether 1/1/0/{port}\n'.format(
                if_id=i, port=100)
        mgr.get_candidate_config.return_value = candidate
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.driver.setup_vlan, "a", "u", "p",
                          "1", 8, self.ports, self.mac)


class TestCFABdriverSetupVlanWithLAG(BaseTestCFABdriver):
    """Test Fujitsu C-Fabric mechanism driver for VLAN configuration.

    """

    def setUp(self):
        cfg.CONF.set_override('pprofile_prefix', "test-", "fujitsu_cfab")
        super(TestCFABdriverSetupVlanWithLAG, self).setUp()
        self.ports = "1/1/0/1,1/1/0/2"

    def test_raises(self):
        mgr = self.driver.mgr
        cfab = self.driver
        for er in [EOFError, EnvironmentError, OSError, select.error]:
            mgr.get_candidate_config.side_effect = er
            self.assertRaises(er,
                              cfab.setup_vlan_with_lag, 'a', 'u', 'p', '1', 8,
                              self.ports, self.mac)
        self.assertEqual(4, mgr.close_session.call_count)

    def test_ifgroup_ether_is_exhauted(self):
        mgr = self.driver.mgr
        candidate = ""
        for i in range(0, 4096):
            candidate += 'ifgroup {if_id} ether 1/1/0/{port}\n'.format(
                if_id=i, port=(i + 1))
        mgr.get_candidate_config.return_value = candidate
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.driver.setup_vlan_with_lag, "a", "u", "p",
                          "1", 8, self.ports, self.mac)

    def test_lag_id_is_exhauted(self):
        mgr = self.driver.mgr
        candidate = ""
        for i in range(1, 200):
            candidate += 'linkaggregation 1 {lag}\n'.format(lag=i)
        mgr.get_candidate_config.return_value = candidate
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.driver.setup_vlan_with_lag, "a", "u", "p",
                          "1", 8, self.ports, self.mac)

    def test_no_preconfig_exist(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = "no_preconfig"
        self.driver.setup_vlan_with_lag("a", "u", "p", "1", 8,
                                        self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('lag', 'add') \
            + cfab_cmd('ifgroup', 'add', lag=True) \
            + cfab_cmd('interface', 'add', ports=self.ports, lag=True) \
            + cfab_cmd('vlan', 'add')
        self.assert_configured(expect)

    def test_exist_lag_id_and_ifgroup(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
ifgroup 1 linkaggregation 1 2
interface 1/1/0/1
    exit
interface 1/1/0/2
    exit
interface 1/1/0/3
    type linkaggregation 1
    exit
interface 1/1/0/4
    type linkaggregation 1
    exit
interface 1/1/0/5
    type linkaggregation 2
    exit
interface 1/1/0/6
    type linkaggregation 2
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
linkaggregation 1 2 cfab port-mode external
linkaggregation 1 2 mode active
linkaggregation 1 2 type endpoint
vfab 1 vlan 8 endpoint untag 0
vfab 1 vlan 16 endpoint untag 1
        """
        self.driver.setup_vlan_with_lag("a", "u", "p", "1", 8,
                                        self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('lag', 'add', lag_id='3') \
            + cfab_cmd('ifgroup', 'add', ifg='2', lag_id='3', lag=True) \
            + cfab_cmd('interface', 'add', lag_id='3',
                       ports=self.ports, lag=True) \
            + cfab_cmd('vlan', 'add', ifg='0,2')
        self.assert_configured(expect)

    def test_illegal_exist_different_lag_id(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
ifgroup 1 linkaggregation 1 2
interface 1/1/0/1
    type linkaggregation 1
    exit
interface 1/1/0/2
    type linkaggregation 2
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
linkaggregation 1 2 cfab port-mode external
linkaggregation 1 2 mode active
linkaggregation 1 2 type endpoint
vfab 1 vlan 8 endpoint untag 0-1
        """
        self.driver.setup_vlan_with_lag("a", "u", "p", "1", 8,
                                        self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'replace', ifg='1') \
            + cfab_cmd('lag', 'delete') \
            + cfab_cmd('lag', 'add', lag_id='3') \
            + cfab_cmd('ifgroup', 'add', ifg='2', lag_id='3', lag=True) \
            + cfab_cmd('interface', 'add', lag_id='3',
                       ports=self.ports, lag=True) \
            + cfab_cmd('vlan', 'add', ifg='1,2')
        self.assert_configured(expect)

    def test_exists_ether_vlan_definition(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1,1/1/0/2
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
interface 1/1/0/2
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
vfab 1 vlan 8 endpoint untag 0
        """
        cfab.setup_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'delete') \
            + cfab_cmd('lag', 'add', lag_id='1') \
            + cfab_cmd('ifgroup', 'add', ifg='1', lag=True) \
            + cfab_cmd('interface', 'add', ports=self.ports, lag=True) \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_reuse_ifgroup(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
ifgroup 1 linkaggregation 1 2
ifgroup 2 linkaggregation 1 3
interface 1/1/0/1
    exit
interface 1/1/0/2
    exit
interface 1/1/0/3
    type linkaggregation 2
    exit
interface 1/1/0/4
    type linkaggregation 2
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
linkaggregation 1 2 cfab port-mode external
linkaggregation 1 2 mode active
linkaggregation 1 2 type endpoint
vfab 1 vlan 16 endpoint untag 1
        """
        self.driver.setup_vlan_with_lag("a", "u", "p", "1", 8,
                                        self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('lag', 'add', lag_id='3') \
            + cfab_cmd('interface', 'add', lag_id='3', ports=self.ports,
                       lag=True) \
            + cfab_cmd('vlan', 'add', ifg='2')
        self.assert_configured(expect)

    def test_exist_lag_id_and_override(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    exit
interface 1/1/0/2
    type linkaggregation 1
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
vfab 1 vlan 100 endpoint untag 0
        """
        self.driver.setup_vlan_with_lag("a", "u", "p", "1", 8,
                                        self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'delete', vlanid=100) \
            + cfab_cmd('lag', 'delete') \
            + cfab_cmd('lag', 'add', lag_id='2') \
            + cfab_cmd('ifgroup', 'add', ifg='1', lag_id='2', lag=True) \
            + cfab_cmd('interface', 'add', lag_id='2',
                       ports=self.ports, lag=True) \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_exist_ifgroup_with_lag_range_and_out_of_range(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1-2
interface 1/1/0/1
    exit
interface 1/1/0/2
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
linkaggregation 1 2 cfab port-mode external
linkaggregation 1 2 mode active
linkaggregation 1 2 type endpoint
        """
        self.driver.setup_vlan_with_lag("a", "u", "p", "1", 8,
                                        self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('lag', 'add', lag_id='3') \
            + cfab_cmd('ifgroup', 'add', ifg='1', lag_id='3', lag=True) \
            + cfab_cmd('interface', 'add', lag_id='3',
                       ports=self.ports, lag=True) \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)

    def test_already_configured_lag_and_vlan(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    exit
interface 1/1/0/2
    type linkaggregation 1
    exit
linkaggregation 1 1 type endpoint
linkaggregation 1 1 mode active
linkaggregation 1 1 cfab port-mode external
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.setup_vlan_with_lag("a", "u", "p", "1", 8,
                                        self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'delete') \
            + cfab_cmd('lag', 'delete') \
            + cfab_cmd('lag', 'add', lag_id='2') \
            + cfab_cmd('ifgroup', 'add', ifg='1', lag_id='2', lag=True) \
            + cfab_cmd('interface', 'add', ports=self.ports,
                       lag_id='2', lag=True) \
            + cfab_cmd('vlan', 'add', ifg='1')
        self.assert_configured(expect)


class TestCFABdriverClearVlan(BaseTestCFABdriver):
    """Test Fujitsu C-Fabric mechanism driver for VLAN configuration."""

    def setUp(self):
        cfg.CONF.set_override('share_pprofile', True, "fujitsu_cfab")
        cfg.CONF.set_override('pprofile_prefix', "test-", "fujitsu_cfab")
        super(TestCFABdriverClearVlan, self).setUp()

    def test_raises(self):
        mgr = self.driver.mgr
        cfab = self.driver
        for er in [EOFError, EnvironmentError, OSError, select.error]:
            mgr.get_candidate_config.side_effect = er
            self.assertRaises(er,
                              cfab.clear_vlan, 'a', 'u', 'p', '1', 8,
                              self.ports, self.mac)
        self.assertEqual(4, mgr.close_session.call_count)

    def test_ifgroup_ether_is_exhauted(self):
        mgr = self.driver.mgr
        candidate = ""
        for i in range(0, 4096):
            candidate += 'ifgroup {if_id} ether 1/1/0/{port}\n'.format(
                if_id=i, port=(i + 1))
        mgr.get_candidate_config.return_value = candidate
        ret = self.driver.clear_vlan("a", "u", "p", "1", 8,
                                     self.ports, self.mac)
        self.assertIsNone(ret)

    def test_clear_with_no_command(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('vlan', 'delete') + cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_clear_vlan_and_assoc(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
pprofile test-1 vlan tag 8
vfab 1 vlan 8 endpoint untag 0
vfab 1 pprofile 0 vsiid mac 00:01:02:03:04:05 test-1
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('pp_assoc', 'delete') \
            + cfab_cmd('vlan', 'delete') \
            + cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_eliminate_own_definition(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
ifgroup 1 ether 1/1/0/2
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
interface 1/1/0/2
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
vfab 1 vlan 8 endpoint untag 0,1
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('vlan', 'replace', ifg='1') \
            + cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_eliminate_own_definition_from_boundary(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/2
ifgroup 1 ether 1/1/0/3
ifgroup 2 ether 1/1/0/4
ifgroup 3 ether 1/1/0/1
ifgroup 4 ether 1/1/0/5
ifgroup 5 ether 1/1/0/6
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
interface 1/1/0/2
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
vfab 1 vlan 8 endpoint untag 0-5
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('vlan', 'replace', ifg='0-2,4-5') \
            + cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_already_cleared_vlan(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    exit
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_already_cleared_vlan_without_interface(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    type endponit
    cfab port-mode external
    lldp mode enable
    exit
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_already_cleared_all_definitions(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
interface 1/1/0/1
    exit
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_exists_different_vlan(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    cfab port-mode external
    type endpoint
    exit
vfab 1 vlan 100 endpoint untag 0
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('vlan', 'delete', vlanid=100) \
            + cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_exists_different_vlans(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
interface 1/1/0/1
    cfab port-mode external
    type endpoint
    exit
vfab 1 vlan 100 endpoint untag 0
vfab 1 vlan 200 endpoint untag 0
vfab 1 vlan 300 endpoint untag 0
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('vlan', 'delete', vlanid=100) \
            + cfab_cmd('vlan', 'delete', vlanid=200) \
            + cfab_cmd('vlan', 'delete', vlanid=300) \
            + cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_exists_different_vlan_with_range(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
ifgroup 1 ether 1/1/0/2
interface 1/1/0/1
    cfab port-mode external
    type endpoint
    exit
vfab 1 vlan 100 endpoint untag 0-1
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('vlan', 'replace', vlanid=100, ifg=1) \
            + cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_exists_lag(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    exit
linkaggregation 1 1 mode active
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 type endpoint
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_exists_lag_without_vlan(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    exit
linkaggregation 1 1 mode active
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 type endpoint
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_exists_lag_without_interface(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    exit
linkaggregation 1 1 mode active
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 type endpoint
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete')
        self.assert_configured(expect)

    def test_illegal_exists_port_range(self):
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1,1/1/0/2
interface 1/1/0/1
    cfab port-mode external
    type endpoint
    exit
interface 1/1/0/2
    cfab port-mode external
    type endpoint
    exit
vfab 1 vlan 8 endpoint untag 0
        """
        self.driver.clear_vlan("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete')
        self.assert_configured(expect)


class TestCFABdriverClearVlanWithLAG(BaseTestCFABdriver):
    """Test Fujitsu C-Fabric mechanism driver for VLAN configuration."""

    def setUp(self):
        cfg.CONF.set_override('share_pprofile', True, "fujitsu_cfab")
        cfg.CONF.set_override('pprofile_prefix', "test-", "fujitsu_cfab")
        super(TestCFABdriverClearVlanWithLAG, self).setUp()
        self.ports = "1/1/0/1,1/1/0/2"

    def test_raises(self):
        mgr = self.driver.mgr
        cfab = self.driver
        for er in [EOFError, EnvironmentError, OSError, select.error]:
            mgr.get_candidate_config.side_effect = er
            self.assertRaises(er,
                              cfab.clear_vlan_with_lag, 'a', 'u', 'p', '1', 8,
                              self.ports, self.mac)
        self.assertEqual(4, mgr.close_session.call_count)

    def test_ifgroup_ether_is_exhauted(self):
        cfab = self.driver
        mgr = self.driver.mgr
        candidate = ""
        for i in range(0, 4096):
            candidate += 'ifgroup {if_id} ether 1/1/0/{port}\n'.format(
                if_id=i, port=(i + 1))
        mgr.get_candidate_config.return_value = candidate
        ret = cfab.clear_vlan_with_lag("a", "u", "p", "1", 8,
                                       self.ports, self.mac)
        self.assertIsNone(ret)

    def test_clear_with_no_command(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    lldp mode enable
    exit
interface 1/1/0/2
    type linkaggregation 1
    lldp mode enable
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
vfab 1 vlan 8 endpoint untag 0
        """
        cfab.clear_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'delete') + cfab_cmd('lag', 'delete')
        self.assert_configured(expect)

    def test_eliminate_own_definition(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
ifgroup 1 linkaggregation 1 2
interface 1/1/0/1
    type linkaggregation 1
    lldp mode enable
    exit
interface 1/1/0/2
    type linkaggregation 1
    lldp mode enable
    exit
interface 1/1/0/3
    type linkaggregation 2
    lldp mode enable
    exit
interface 1/1/0/4
    type linkaggregation 2
    lldp mode enable
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
linkaggregation 1 2 cfab port-mode external
linkaggregation 1 2 mode active
linkaggregation 1 2 type endpoint
vfab 1 vlan 8 endpoint untag 0,1
        """
        cfab.clear_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'replace', ifg='1') \
            + cfab_cmd('lag', 'delete')
        self.assert_configured(expect)

    def test_already_cleared_only_interface(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    lldp mode enable
    exit
interface 1/1/0/2
    lldp mode enable
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
vfab 1 vlan 8 endpoint untag 0
        """
        cfab.clear_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports)
        self.assert_configured(expect)

    def test_already_cleared_only_vlan(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    lldp mode enable
    exit
interface 1/1/0/2
    type linkaggregation 1
    lldp mode enable
    exit
linkaggregation 1 1 cfab port-mode external
linkaggregation 1 1 mode active
linkaggregation 1 1 type endpoint
        """
        cfab.clear_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('lag', 'delete')
        self.assert_configured(expect)

    def test_already_cleared_lag_definition(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 linkaggregation 1 1
interface 1/1/0/1
    type linkaggregation 1
    lldp mode enable
    exit
interface 1/1/0/2
    type linkaggregation 1
    lldp mode enable
    exit
vfab 1 vlan 8 endpoint untag 0
        """
        cfab.clear_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports) \
            + cfab_cmd('vlan', 'delete') \
            + cfab_cmd('lag', 'delete')
        self.assert_configured(expect)

    def test_already_cleared_all_definitions(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
interface 1/1/0/1
    exit
        """
        cfab.clear_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports)
        self.assert_configured(expect)

    def test_exists_ether_vlan(self):
        cfab = self.driver
        mgr = self.driver.mgr
        mgr.get_candidate_config.return_value = """
ifgroup 0 ether 1/1/0/1
ifgroup 1 ether 1/1/0/2
interface 1/1/0/1
    type endpoint
    cfab port-mode external
    exit
interface 1/1/0/2
    type endpoint
    cfab port-mode external
    exit
vfab 1 vlan 8 endpoint untag 0-1
        """
        cfab.clear_vlan_with_lag("a", "u", "p", "1", 8, self.ports, self.mac)
        mgr.connect.assert_called_once_with("a", "u", "p")
        mgr.get_candidate_config.assert_called_once_with()
        expect = cfab_cmd('interface', 'delete', ports=self.ports)
        self.assert_configured(expect)


class TestCFABdriverPrivateMethods(BaseTestCFABdriver):
    """Test Fujitsu C-Fabric mechanism driver with private methods.

    This class is for illegal case tests.

    """

    def setUp(self):
        cfg.CONF.set_override('share_pprofile', True, "fujitsu_cfab")
        cfg.CONF.set_override('pprofile_prefix', "test-", "fujitsu_cfab")
        super(TestCFABdriverPrivateMethods, self).setUp()

    def test_is_ifgroup_included_between_range(self):
        ifgroup_id = 2
        ifgroups = '1-3'
        self.assertTrue(cfabdriver._is_ifgroup_included(ifgroup_id, ifgroups))


def cfab_cmd(target, op, vfab_id='1', vlanid=8, pp_name='test-1', ppid='0',
             lag=False, ports='1/1/0/1', domain_id='1', lag_id='1', ifg='0'):
    """Expected result for C-Fabric commands via operations.

    @param target  A string of target definition name. Following targets are
                   available:
                       'interface', 'lag', 'vlan', 'ifgroup', 'ifgroup_lag'
                       'assoc', 'pp_assoc'
    @param op      A string of operation. Following operations are available:
                       'add', 'replace', 'delete'
    @return  ret   A list of string object

    """
    ret = []
    if target is 'interface':
        ret.append("interface range {ports}".format(ports=ports))
        if op is 'add' or op is 'replace':
            if lag:
                ret.append('type linkaggregation {lag}'.format(lag=lag_id))
            else:
                ret.append("cfab port-mode external")
                ret.append("type endpoint")
        if op is 'delete':
            ret.append('no type')
            ret.append("no cfab port-mode")
        ret.append('exit')

    elif target is 'lag':
        lag = 'linkaggregation {dom} {lag}'.format(lag=lag_id, dom=domain_id)
        if op is 'add' or op is 'replace':
            ret.append(lag + ' cfab port-mode external')
            ret.append(lag + ' mode active')
            ret.append(lag + ' type endpoint')
        if op is 'delete':
            ret.append('no ' + lag + ' cfab port-mode')
            ret.append('no ' + lag + ' mode')
            ret.append('no ' + lag + ' type')

    elif target is 'vlan':
        if op is 'add' or op is 'replace':
            ret.append("vfab {vfab} vlan {vid} endpoint untag {ifg}".format(
                vfab=vfab_id, vid=vlanid, ifg=ifg))
        if op is 'delete':
            ret.append("no vfab {vfab} vlan {vid} endpoint untag".format(
                vfab=vfab_id, vid=vlanid))

    elif target is 'ifgroup':
        if op is 'add':
            if lag:
                ret.append("ifgroup {ifg} linkaggregation {dom} {lag}".format(
                    ifg=ifg, dom=domain_id, lag=lag_id))
            else:
                ret.append('ifgroup {ifg} ether {p}'.format(ifg=ifg, p=ports))

    elif target is 'assoc':
        if op is 'delete':
            ret.append("no pprofile {pp_name}".format(pp_name=pp_name))

    elif target is 'pp_assoc':
        if op is 'delete':
            ret.append("no vfab {vfab} pprofile {ppid}".format(
                vfab=vfab_id, ppid=ppid))
            ret.append("no pprofile {pp_name}".format(pp_name=pp_name))
    else:
        raise 'Illegal target(%s) is specified' % target
    return ret



# Copyright 2015-2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import os

from networking_fujitsu.ml2.cfab import cfabdriver
from networking_fujitsu.ml2.cfab import mech_cfab
from networking_fujitsu.tests.unit.ml2.common import helper
from neutron.plugins.ml2 import config as ml2_config
from neutron.tests.unit.plugins.ml2 import test_plugin as test_ml2_plugin

USERNAME = 'admin'
PASSWORD = 'admin'
ADDRESS = '192.168.100.1'
PHYSICAL_NET = ["physnet1:1"]


class TestFujitsuMechDriverV2(test_ml2_plugin.Ml2PluginV2TestCase):
    """Test Fujitsu mechanism driver.

    If the environment variable OS_FUJITSU_CFAB_ADDRESS is defined, tests will
    be performed using the C-Fabric driver, and a telnet connection to the
    address OS_FUJITSU_CFAB_ADDRESS and the port OS_FUJITSU_CFAB_PORT (defaults
    to 23) will be established. The username will be taken from
    OS_FUJITSU_CFAB_USERNAME (defaults to "admin") and the password will be
    taken from OS_FUJITSU_CFAB_PASSWORD (defaults to "password").
    If the environment variable OS_FUJITSU_CFAB_ADDRESS is NOT defined, tests
    will be performed using the mock driver instead of the C-Fabric driver, and
    no real telnet connection will be used.
    """

    _mechanism_drivers = ['fujitsu_cfab']
    # test_create_router_port_and_fail_create_postcommit:
    #     This one is mocked 'fake_driver' only. So, our plugin's instance
    #     hasn't mocked and fail definitely. Therefore, skips this test.
    _skip = ["test_create_router_port_and_fail_create_postcommit"]

    def setUp(self):

        ml2_config.cfg.CONF.set_override(
            'tenant_network_types', ['vlan'], 'ml2')
        ml2_config.cfg.CONF.set_override(
            'type_drivers', ['vlan'], 'ml2')

        if self._testMethodName in self._skip:
            self.skipTest("This test has already verified at neutron's test.")

        address = os.environ.get('OS_FUJITSU_CFAB_ADDRESS')
        if address:
            ml2_fujitsu_opts = {
                'username': os.environ.get('OS_FUJITSU_CFAB_USERNAME') or
                "admin",
                'password': os.environ.get('OS_FUJITSU_CFAB_PASSWORD') or
                "password",
                'address': address,
                'physical_networks': ["physnet1:1", "physnet2:2"],
                'pprofile_prefix': "test-"}

            for opt, val in ml2_fujitsu_opts.items():
                ml2_config.cfg.CONF.set_override(opt, val, "fujitsu_cfab")

            try:
                cfabdriver.TELNET_PORT = int(
                    os.environ['OS_FUJITSU_CFAB_PORT'])
            except KeyError:
                pass
            super(TestFujitsuMechDriverV2, self).setUp()
        else:
            def mocked_initialize(self):
                self._switch = {'address': "", 'username': "", 'password': ""}
                self._driver = mock.MagicMock()
                self._physical_networks = {'physnet1': "1", 'physnet2': "2"}

            with mock.patch.object(mech_cfab.CFABMechanismDriver,
                                   'initialize', new=mocked_initialize):
                super(TestFujitsuMechDriverV2, self).setUp()


class TestFujitsuMechDriverPortsV2(test_ml2_plugin.TestMl2PortsV2,
                                   TestFujitsuMechDriverV2):
    pass


class TestMechCFABIsSupported(helper.FujitsuMechanismHelper):

    def test_is_supported_type_vlan(self):
        ctx = self.prepare_dummy_context('network')
        network = ctx.current
        self.assertTrue(mech_cfab.is_supported(network))

    def test_is_supported_not_support_type_flat(self):
        ctx = self.prepare_dummy_context('network', net_type='flat')
        network = ctx.current
        self.assertFalse(mech_cfab.is_supported(network))

    def test_is_supported_not_support_type_vxlan(self):
        ctx = self.prepare_dummy_context('network', net_type='vxlan')
        network = ctx.current
        self.assertFalse(mech_cfab.is_supported(network))

    def test_is_supported_illegal_seg_id_is_none(self):
        ctx = self.prepare_dummy_context('network', net_type='vlan')
        network = ctx.current
        ctx.current['provider:segmentation_id'] = None
        self.assertFalse(mech_cfab.is_supported(network))

    def test_is_supported_illegal_seg_id_is_not_defined(self):
        ctx = self.prepare_dummy_context('network', net_type='vlan')
        network = ctx.current
        del network['provider:segmentation_id']
        self.assertFalse(mech_cfab.is_supported(network))


class TestFujitsuMechDriverBaremetalPortsV2(helper.FujitsuMechanismHelper):

    def setUp(self):
        ml2_fujitsu_opts = {
            'username': USERNAME,
            'password': PASSWORD,
            'address': ADDRESS,
            'physical_networks': PHYSICAL_NET,
            'pprofile_prefix': 'test-'
        }
        for opt, val in ml2_fujitsu_opts.items():
            ml2_config.cfg.CONF.set_override(opt, val, "fujitsu_cfab")
        self.mech = mech_cfab.CFABMechanismDriver()
        self.mech._driver = mock.Mock()
        super(TestFujitsuMechDriverBaremetalPortsV2, self).setUp()

    def test_create_port(self):
        ctx = self.prepare_dummy_context()
        self.mech.create_port_postcommit(ctx)
        self.mech._driver.associate_mac_to_network.assert_not_called()

    def test_bind_port_with_single(self):
        ctx = self.prepare_dummy_context()
        self.mech.bind_port(ctx)
        params = params_for_setup_vlan(ctx.current)
        self.mech._driver.setup_vlan.assert_called_once_with(
            params['address'],
            params['username'],
            params['password'],
            params['vfab_id'],
            params['vlanid'],
            params['ports'],
            params['mac_address'],
        )

    def test_bind_port_with_lag(self):
        ctx = self.prepare_dummy_context(nic='lag')
        self.mech.bind_port(ctx)
        params = params_for_setup_vlan(ctx.current, lag=True)
        self.mech._driver.setup_vlan_with_lag.assert_called_once_with(
            params['address'],
            params['username'],
            params['password'],
            params['vfab_id'],
            params['vlanid'],
            params['ports'],
            params['mac_address'],
        )

    def test_bind_port_with_mlag(self):
        ctx = self.prepare_dummy_context(nic='mlag')
        self.mech.bind_port(ctx)
        params = params_for_setup_vlan(ctx.current, lag=True)
        self.mech._driver.setup_vlan_with_lag.assert_called_once_with(
            params['address'],
            params['username'],
            params['password'],
            params['vfab_id'],
            params['vlanid'],
            params['ports'],
            params['mac_address'],
        )

    def test_update_port(self):
        ctx = self.prepare_dummy_context()
        self.mech.update_port_postcommit(ctx)
        self.mech._driver.associate_mac_to_network.assert_not_called()

    def test_delete_port_with_single(self):
        ctx = self.prepare_dummy_context(nic='single')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_setup_vlan(ctx.current)
        self.mech._driver.clear_vlan.assert_called_with(
            params['address'],
            params['username'],
            params['password'],
            params['vfab_id'],
            params['vlanid'],
            params['ports'],
            params['mac_address'],
        )
        self.mech._driver.dissociate_mac_from_network.assert_not_called()

    def test_delete_with_lag(self):
        ctx = self.prepare_dummy_context(nic='lag')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_setup_vlan(ctx.current, lag=True)
        self.mech._driver.clear_vlan_with_lag.assert_called_with(
            params['address'],
            params['username'],
            params['password'],
            params['vfab_id'],
            params['vlanid'],
            params['ports'],
            params['mac_address'],
        )
        self.mech._driver.dissociate_mac_from_network.assert_not_called()

    def test_delete_with_mlag(self):
        ctx = self.prepare_dummy_context(nic='mlag')
        self.mech.delete_port_postcommit(ctx)
        params = params_for_setup_vlan(ctx.current, lag=True)
        self.mech._driver.clear_vlan_with_lag.assert_called_with(
            params['address'],
            params['username'],
            params['password'],
            params['vfab_id'],
            params['vlanid'],
            params['ports'],
            params['mac_address'],
        )
        self.mech._driver.dissociate_mac_from_network.assert_not_called()


def params_for_setup_vlan(port, lag=False):
    lli = port['binding:profile']['local_link_information']
    return {
        'address': ADDRESS,
        'username': USERNAME,
        'password': PASSWORD,
        'vfab_id': '1',
        'vlanid': 1111,
        'ports': ','.join(p['port_id'] for p in lli),
        'mac_address': port['mac_address'],
        'lag': lag
    }



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import MagicMock
from networking_fujitsu.ml2.ism import ism_base
from neutron_lib.api.definitions import portbindings
from oslo_utils import uuidutils
import random

VERSION = 'ism/api/v1'
APIS = {'get_node': 'ism/api/v1/core/nodes/%(node)s',
        'list_ports': 'ism/api/v1/topology/nodes/%(node)s/%(port)s',
        'get_port': 'ism/api/v1/topology/nodes/%(node)s/ports/%(port)s',
        'get_vlan': 'ism/api/v1/topology/nodes/%(node)s/ports/%(port)s/vlan',
        'patch_vlan': 'ism/api/v1/topology/nodes/%(node)s/ports/%(port)s/vlan',
        'get_vxlan_with_port': '',
        }

RES = {'get_node': {'IsmBody': {
                    'Node': {
                        'TimeStampInfo': {
                            'Delete': '',
                            'Update': '2015-04-14T15:30:45.250Z',
                            'Register': '2015-04-14T15:30:45.250Z'},
                        'IpVersion': 'V4',
                        'IpAddress': '192.168.123.15',
                        'NodeId': '%(node)s',
                        'Name': 'blade-server1-switch1',
                        'Model': 'PY CB Eth Switch/IBP 10Gb 18/8',
                        'Status': 'Normal',
                        'AlarmStatus': 'Normal',
                        'Type': 'server',
                        'WebUrl': 'https://192.168.123.15',
                        'Description': ''},
                    'SchemaType': 'http://get_node.json'
                    }},
       'list_ports': {'IsmBody': {
                      'PortInfo': [{
                          'NeighborPort': [{
                              'PortName': 'ether 2',
                              'NodeId': '2'}],
                          'PortNumber': '2',
                          'ControllerName': '',
                          'Category': 'LAN',
                          'PortName': 'ether 2'}, {
                          'NeighborPort': [{
                              'PortName': 'eth0',
                              'NodeId': '6'}],
                          'PortNumber': '3',
                          'ControllerName': '',
                          'Category': 'LAN',
                          'PortName': 'ether 3'}, {
                          'NeighborPort': [],
                          'PortNumber': '4',
                          'ControllerName': '',
                          'Category': 'LAN',
                          'PortName': 'ether 4'}],
                      'SchemaType': 'http://list_ports.json'}},
       'get_port': {'IsmBody': {
                    'ControllerName': '',
                    'DetailedInfo': {
                        'IpAddressSet': {
                            'V4': '',
                            'V6': ''},
                        'MacAddress': '00:0C:29:31:88:EF',
                        'PortAddress': '',
                        'Slot': '',
                        'Speed': {
                            'Unit': 'Mpbs',
                            'Value': 1000
                        },
                        'Wwpn': ''},
                    'NeighborPort': [{
                        'NodeId': '2',
                        'PortName': 'ether 2'}],
                    'PortCategory': 'LAN',
                    'PortName': 'ether 1',
                    'PortNumber': '1',
                    'SchemaType': 'http://get_port.json'}},
       'get_vlan': {'IsmBody': {
                    'TaggedVLanId': '100,200,40-50',
                    'UntaggedVLanId': '300',
                    'SchemaType': 'http://get_vlan.json'}},
       'patch_vlan': {'IsmBody': {
                      'TaggedVLanId': '100,200,40-50',
                      'UntaggedVLanId': '300',
                      'SchemaType': 'http://get_vlan.json'}},
       'get_vxlan_with_port': '',
       }


class Factory(object):
    def __init__(self):
        pass


class IsmMockFactory(Factory):
    def __init__(self, network_type, seg_id):
        ism_base.IsmVlanBase = MagicMock()
        ism_base.IsmVxlanBase = MagicMock()
        self.initialize(network_type, seg_id)

    def initialize(self, network_type, seg_id):
        self.create_ism_base = MagicMock()
        if (network_type == 'vlan'):
            self.create_ism_base.return_value = ism_base.IsmVlanBase
        else:
            self.create_ism_base.return_value = ism_base.IsmVxlanBase

    def _ret_ism_info(self, params=None):
        params = params
        # TODO(yushiro) prepare response body from ISM
        return {'IsmBody': 'dummy'}

    def _req_param(self, tags=None, untags=None):
        tagged_vlan = tags or ''
        untagged_vlan = untags or '200'
        return {'TaggedVLanId': tagged_vlan, 'UntaggedVLanId': untagged_vlan}

    def _header_and_body(self, headers=None, body={}):
        headers = headers or {'sample': 'of', 'http': 'header'}
        body = body or {'IsmBody': {'key': 'value'}}
        return headers, body


class NeutronFactory(Factory):
    def __init__(self):
        self.tenant_id = '7e02058126cc4950b75f9970368ba177'
        # TODO(yushiro): delete if Ironic-Neutron Integration patch is merged.
        portbindings.VNIC_BAREMETAL = 'baremetal'

    def build_network(self, params={}):
        network_type = params.get('provider:network_type') or 'vlan'
        seg_id = params.get('provider:segmentation_id') or 1234
        return {'id': uuidutils.generate_uuid(),
                'name': 'test-network',
                'subnets': [],
                'admin_state_up': True,
                'status': 'ACTIVE',
                'tenant_id': self.tenant_id,
                'shared': False,
                'provider:network_type': network_type,
                'provider:segmentation_id': seg_id,
                }

    def build_port(self, network, params={}):
        vnic_type = (params.get('binding:vnic_type') or
                     portbindings.VNIC_BAREMETAL)
        local_link_info = self.__default_link_local_info()
        if hasattr(params, 'local_link_information'):
            local_link_info = params['local_link_information']
        return {'id': uuidutils.generate_uuid(),
                'name': 'test-port',
                'network_id': network['id'],
                'admin_state_up': True,
                'mac_address': self.__randomMAC(),
                'fixed_ips': None,
                'device_id': '',
                'device_owner': '',
                'tenant_id': network['tenant_id'],
                'status': 'ACTIVE',
                'binding:vnic_type': vnic_type,
                'binding:profile': {'local_link_information': local_link_info}
                }

    def __default_link_local_info(self):
        return [{'switch_id': '5',
                 'switch_info': 'switch_1',
                 'port_id': 'ether 1',
                 }]

    def __randomMAC(self):
        _mac = [0x00, 0x16, 0x3e,
                random.randint(0x00, 0x7f),
                random.randint(0x00, 0xff),
                random.randint(0x00, 0xff)]
        mac = ':'.join(map(lambda x: "%02x" % x, _mac))
        return mac


class DummyContext(object):

    def __init__(self, mock_net, mock_port):
        self.network = mock_net
        self.port = mock_port
        self.current = mock_port



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from neutron.plugins.ml2.drivers import type_tunnel
from neutron.plugins.ml2 import managers
from neutron.plugins.ml2 import rpc
from neutron.tests import base

from networking_fujitsu.ml2.common import tunnel_caller


class TestTunnelCaller(base.BaseTestCase):

    def setUp(self):
        super(TestTunnelCaller, self).setUp()
        mock.patch.object(managers, 'TypeManager').start()
        self.tunnel_caller = tunnel_caller.TunnelCaller()
        self.context = mock.MagicMock()

    def test_trigger_tunnel_sync(self):
        with mock.patch.object(rpc, 'RpcCallbacks'), \
                mock.patch.object(type_tunnel.TunnelRpcCallbackMixin,
                                  'tunnel_sync') as mock_tunnel_sync:
            self.tunnel_caller.trigger_tunnel_sync(self.context, 'fake_ip')
            mock_tunnel_sync.assert_called_with(
                self.context, tunnel_ip='fake_ip', tunnel_type='vxlan')



# Copyright 2015-2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LW
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net
from neutron_lib.plugins.ml2 import api
from oslo_log import log as logging
import re

LOG = logging.getLogger(__name__)
RANGE_DEFINITION = re.compile(r'(\d+)-(\d+)')
_SUPPORTED_NET_TYPES = ['vlan']


def eliminate_val(source, reject):
    """Eliminate specified value from range value.

    ex. source='1,2,3-10', reject=[1], result: '2,3-10'
    :param source: a string of range definition separated with ","
    :type source: string
    :param reject: a list of integer to reject. ex. [1, 2]
    :type reject: list of integer

    :returns: a string separated with "," like an example
    :rtype: string
    """
    if source is None:
        return
    values = [str(i) for i in reject]
    rejected = source.split(',')
    found = False
    for val in values:
        for target in rejected:
            m = RANGE_DEFINITION.match(target)
            if m:
                low = m.group(1)
                high = m.group(2)
                if val in target:
                    rejected.remove(target)
                    # matches the lowest one
                    if (val == low):
                        # Case: source is "1-2" and target is "1"
                        if ((int(val) + 1) == int(high)):
                            rejected.append(high)
                        else:
                            rejected.append(str(int(val) + 1) + "-" + high)
                            found = True
                            break
                    # matches the highest one
                    else:
                        # Ex. source is "1-2" and target is "2"
                        if ((int(val) - 1) == int(low)):
                            rejected.append(low)
                        else:
                            rejected.append(low + "-" + str(int(val) - 1))
                        found = True
                        break
                # matches between lower one and higher one
                elif (int(low) < int(val) and int(val) < int(high)):
                    rejected.remove(target)
                    # Ex. source is "1-n" and target is "2"
                    if ((int(val) - 1) == int(low)):
                        rejected.append(low)
                        # Ex. source is "1-3" and target is "2"
                        if ((int(val) + 1) == int(high)):
                            rejected.append(high)
                        # Ex. source is "1-4" and target is "2"
                        else:
                            rejected.append(str(int(val) + 1) + "-" + high)
                    # Ex. source is "n-5" and target is "4"(n is NOT "3")
                    elif ((int(val) + 1) == int(high)):
                        rejected.append(high)
                        rejected.append(low + "-" + str(int(val) - 1))
                    # Ex. source is "1-5" and target is "3"
                    else:
                        rejected.append(low + "-" + str(int(val) - 1))
                        rejected.append(str(int(val) + 1) + "-" + high)
                    found = True
                    break
            # source is not defined with range "-"
            elif val == target:
                rejected.remove(target)
                found = True
                break
    if not found:
        LOG.debug(_('Reject target doesn\'t exist.'))
    return ','.join(rejected)


def _get_provider_attribute(net, net_attr):
    if isinstance(net, dict):
        segment = net['segments'][0] if net.get('segments', None) else net
    else:
        segment = net.network_segments[0]
    try:
        api_attr = getattr(api, net_attr)
        provider_attr = getattr(provider_net, net_attr)
    except AttributeError as e:
        msg = _('Attribute(%s) not found. A version of neutron-lib seems '
                'not to match suitable one. Please check it.'), net_attr
        LOG.critical(msg)
        raise e
    else:
        return segment.get(api_attr, segment.get(provider_attr, None))


def get_network_type(network):
    """Get provider:network_type

    :param network: NetworkContext or network dict
    :type network: NetworkContext or dict

    :returns network_type: 'vlan', 'vxlan', 'local', 'flat' and 'geneve'
    :rtype: string
    """
    return _get_provider_attribute(network, 'NETWORK_TYPE')


def get_segmentation_id(network):
    """Get provider:segmentation_id

    :param network: NetworkContext or network dict
    :type network: NetworkContext or dict

    :returns segmentation_id: VLANID('vlan') or VNI('vxlan')
    :rtype: integer
    """
    return _get_provider_attribute(network, 'SEGMENTATION_ID')


def get_physical_network(network):
    """Get provider:physical_network

    :param network: NetworkContext or network dict
    :type network: NetworkContext or dict

    :returns physical_network: physical network name for the network
    :rtype: string
    """
    return _get_provider_attribute(network, 'PHYSICAL_NETWORK')


def get_physical_connectivity(port):
    """Get local_link_information from specified port.

    :param port: a dictionary of neutron port
    :type port: dictionary

    :returns lli: "local_link_information" as follows
                 [{"switch_id": "MAC_of_switch", "port_id": "1/1/0/1",
                  "switch_info": "switch_name"}]
                 If all of parameter specified, returns above format otherwise
                 empty list
    :rtype: list of dict or empty list
    """

    # TODO(yushiro) replace following characters to constant value
    lli = port['binding:profile'].get("local_link_information", [])
    if not lli:
        return []
    is_all_specified = True
    for i in lli:
        if not (i.get('switch_id') and i.get('port_id') and
                i.get('switch_info')):
            is_all_specified = False
    if is_all_specified:
        return lli
    LOG.warning(_LW("Some physical network param is missing:%s"), lli)
    return []


def is_baremetal(port):
    """Judge a specified port is for baremetal or not.

    :param port: a dictionary of neutron port
    :type port: dictionary

    :returns: True(vnic_type is baremetal) or False(otherwise)
    :rtype: boolean
    """

    vnic_type = port.get(portbindings.VNIC_TYPE, portbindings.VNIC_NORMAL)
    return (vnic_type == portbindings.VNIC_BAREMETAL)


def is_lag(local_link_information):
    """Judge a specified port param is for LAG(linkaggregation) or not.

    :param local_link_information: physical connectivity information
    :type local_link_information: list of dict

    :returns: True(mode is LAG) or False(otherwise)
    :rtype: boolean
    """

    return (len(local_link_information) > 1)



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import httplib2
import json
import os.path
import re
import subprocess
import urllib

from six.moves import http_client as httplib

from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

ML2_ISM_GROUP = "fujitsu_ism"
ML2_FUJITSU_ISM = [
    cfg.StrOpt('address', default='furukawa-ism',
               help=("The IP address or hostname of the ISM.")),
    cfg.StrOpt('username', default='admin',
               help=("The ISM username to use.")),
    cfg.StrOpt('password', default='admin', secret=True,
               help=("The ISM password to use.")),
    cfg.StrOpt('port', default='25566',
               help=("The port number of the ISM.")),
    cfg.StrOpt('certificate_authority',
               default='/etc/neutron/plugins/ml2/fujitsu/server.crt',
               help=("The certification authority for ISM.")),
    cfg.StrOpt('timeout', default=30,
               help=("The API timeout value for ISM."))]

cfg.CONF.register_opts(ML2_FUJITSU_ISM, ML2_ISM_GROUP)

STATUS_SUCCESS = [httplib.OK,  # 200
                  httplib.CREATED,  # 201
                  httplib.ACCEPTED,  # 202
                  httplib.NON_AUTHORITATIVE_INFORMATION,  # 203
                  httplib.NO_CONTENT,  # 204
                  httplib.RESET_CONTENT,  # 205
                  httplib.PARTIAL_CONTENT,  # 206
                  httplib.MULTI_STATUS,  # 207
                  httplib.IM_USED]  # 226

TIMEOUT = int(cfg.CONF.fujitsu_ism.timeout)
VLAN_PATTERN = '^%s$|%s[,-]|[,-]%s[,-]|[,-]%s'
ISM_VERSION = 'v1/'
ISM_URI_PREFIX = '/ism/api/' + ISM_VERSION
ISM_URI_NODES = ISM_URI_PREFIX + 'topology/nodes/%(node)s/'
ISM_URI_PORTS = ISM_URI_PREFIX + 'topology/nodes/%(node)s/ports/%(port)s/'
ISM_BODY = 'IsmBody'
ISM_AUTH_COMMAND = 'echo "%(user)s:%(pass)s" | openssl enc -aes-256-cbc ' \
                   '-e -base64 -pass pass:%(addr)s'
HTTPS = 'https://%s:%s' % (cfg.CONF.fujitsu_ism.address,
                           cfg.CONF.fujitsu_ism.port)


class IsmBase(object):

    def __init__(self, network_type=None, segmentation_id=None):
        self.address = cfg.CONF.fujitsu_ism.address
        self.username = cfg.CONF.fujitsu_ism.username
        self.password = cfg.CONF.fujitsu_ism.password
        self.port = cfg.CONF.fujitsu_ism.port
        self.auth_header = self.generate_auth_header()
        self.ca_cert = cfg.CONF.fujitsu_ism.certificate_authority
        self.network_type = network_type
        # TODO(yushiro): research segmentation_id's type
        self.segmentation_id = segmentation_id

    def do_request(self, uri, method='GET', body={}, headers=None,
                   keyfile=None, certfile=None, ca_certs=None,
                   timeout=TIMEOUT):
        # TODO(yushiro) Confirm ISM specification about certs file

        if ca_certs:
            if not os.path.isfile(ca_certs):
                raise FujitsuIsmException("ca_cert(%s) not found." % ca_certs)

        https = httplib2.Http(ca_certs=ca_certs, timeout=timeout)
        https.force_exception_to_status_code = True
        if body is not None:
            body = urllib.urlencode(body)
        LOG.info("request URI=%s, method=%s" % (uri, method))
        LOG.info("headers=%s, ca_certs=%s" % (headers, ca_certs))
        LOG.info("request body: %s" % body)

        # through request
        try:
            res_headers, res_body = https.request(uri,
                                                  method,
                                                  headers=headers,
                                                  body=body)
            status = int(res_headers.get('status'))
            if (status in STATUS_SUCCESS):
                LOG.info("ISM response: %s" % res_headers)
                LOG.debug("ISM res_body: %s" % res_body)
                return res_headers, json.loads(res_body)
            else:
                LOG.warning("ISM res_headers: %s" % res_headers)
                LOG.warning("ISM res_body: %s" % res_body)
                raise FujitsuIsmException(res_body)
        except Exception as er:
            LOG.exception("Unknown ISM error. detail=%s" % er)
            raise FujitsuIsmException(er)

    def generate_auth_header(self):
        '''Generate authentication header for ISM'''
        auth_command = ISM_AUTH_COMMAND % {'user': self.username,
                                           'pass': self.password,
                                           'addr': self.address}
        try:
            encrypted = subprocess.check_output(auth_command, shell=True)
        except subprocess.CalledProcessError as er:
            LOG.exception("ISMPlugin configuration error.(%s)" % type(er))
            LOG.exception("detail: %s" % er)
            raise FujitsuIsmException(er)
        return 'ISM: ' + encrypted

    def generate_req_param_for_port(self, phy_info):
        '''Generate request parameter for port'''
        pass

    def generate_req_param_for_switch(self, phy_info):
        '''Generate request parameter for switch'''
        pass

    def setup_for_switch(self, phy_info):
        '''Setup switch configuration using ISM API'''
        pass

    def setup_for_port(self, req_body, phy_info):
        '''Setup port configuration using ISM API'''
        pass

    def get_current_config(self, params):
        '''Get current configuration using ISM API'''
        pass

    def trim_ism_body(self, body):
        '''Trim ISM body for request.

        ISM's response body includes unnecessary attribute named "SchemaType".
        '''
        trimmed = dict(body).get(ISM_BODY)
        del trimmed['SchemaType']
        return trimmed


class IsmVlanBase(IsmBase):

    def __init__(self, network_type=None, segmentation_id=None):
        super(IsmVlanBase, self).__init__(network_type, segmentation_id)

    def get_current_config(self, params):
        '''Get current VLAN configuration of specified switch/port

        :params:  Physical connections information from Ironic
        :returns: Response body of VLAN configurations with specified switch
                  and port
        '''
        # TODO(yushiro) get ISM nodeid from switch_id(MAC)
        nodeid = urllib.quote(str(params['switch_id']))
        portname = urllib.quote(str(params['port_id']))
        _uri = HTTPS + ISM_URI_PORTS % {'node': nodeid, 'port': portname}
        uri = _uri + self.network_type
        headers = {'Content-Type': 'application/json',
                   'Authorization': self.auth_header}
        headers, res_body = self.do_request(uri,
                                            method='GET',
                                            body=None,
                                            headers=headers,
                                            ca_certs=self.ca_cert)
        return res_body

    def setup_for_port(self, req_body, params):
        '''Set specified VLANID with specified switch-port

        '''
        # TODO(yushiro) get ISM nodeid from switch_id(MAC)
        nodeid = urllib.quote(str(params['switch_id']))
        portname = urllib.quote(str(params['port_id']))
        _uri = HTTPS + ISM_URI_PORTS % {'node': nodeid, 'port': portname}
        uri = _uri + self.network_type
        req_headers = {'Content-Type': 'application/json',
                       'Authorization': self.auth_header}
        headers, res_body = self.do_request(uri,
                                            method='PATCH',
                                            body=req_body,
                                            headers=req_headers,
                                            ca_certs=self.ca_cert)
        return headers, res_body

    def generate_req_param_for_port(self, body,
                                    vlan_type="UntaggedVLanId"):
        '''Generate request parameter for PATCH VLAN '''
        customized = self.trim_ism_body(body)
        vlan_param = customized.get(vlan_type)
        vlanid = self.segmentation_id

        if not self.__judge_vlanid_is_included(vlan_param, vlanid, vlan_type):
            if (vlan_type == 'UntaggedVLanId'):
                modify_param = self.segmentation_id
            if (vlan_type == 'TaggedVLanId'):
                modify_param = vlan_param + ',' + self.segmentation_id

        customized[vlan_type] = modify_param
        return customized

    def __judge_vlanid_is_included(self, vlan_param, vlanid, vlan_type):
        _pattern = VLAN_PATTERN % (vlanid, vlanid, vlanid, vlanid)
        pattern = r'%s' % _pattern
        if (re.match(pattern, vlan_param)):
            return True
        elif (vlan_type == 'UntaggedVLanId'):
            return False

        # Taggedvlanid case
        boundary = r'(\d+)-(\d+)'
        matched_list = re.findall(boundary, vlan_param)
        for vlan in matched_list:
            if (int(vlan[0]) <= vlanid and vlanid <= int(vlan[1])):
                return True
        return False


class IsmVxlanBase(IsmBase):

    def __init__(self, context, network_type=None, segmentation_id=None):
        super(IsmVxlanBase, self).__init__(network_type, segmentation_id)

        def get_current_config(self, params):
            '''Get current VXLAN configuration of specified switch

            '''
            pass

        def generate_req_param_for_switch(self, phy_info):
            '''Generate request parameter for PATCH VXLAN node

            '''
            pass

        def generate_req_param_for_port(self, phy_info):
            '''Generate request parameter for PATCH VLAN

            '''
            pass

        def setup_for_switch(self, phy_info):
            pass

        def setup_for_port(self, req_body, phy_info):
            pass


class FujitsuIsmException(Exception):
    '''FujitsuIsmException

    '''



# Copyright 2015-2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import mock

from neutron_lib.api.definitions import portbindings

from collections import defaultdict
from networking_fujitsu.ml2.common import utils
from neutron.tests import base


class FujitsuCommonUtilsTestCase(base.BaseTestCase):
    """Base class to test Fujitsu Common Utils"""

    def setUp(self):
        super(FujitsuCommonUtilsTestCase, self).setUp()
        self.net = mock.MagicMock()

        self.net.network_segments = [{
            "provider:network_type": "vlan",
            "provider:physical_network": 'physnet1',
            "provider:segmentation_id": 100
        }]
        self.port = defaultdict(lambda: {})
        self.port[portbindings.VNIC_TYPE] = portbindings.VNIC_BAREMETAL
        self.lli = [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "1/1/0/1",
                     "switch_info": "CFX2000R"},
                    {"switch_id": "00:00:4c:ee:e5:40", "port_id": "1/1/0/2",
                     "switch_info": "CFX2000R"}]
        self.port['binding:profile']['local_link_information'] = self.lli


class TestEliminateVal(FujitsuCommonUtilsTestCase):
    """Test class for eliminate_val"""

    def test_definition_is_None(self):
        definition = None
        target = [1]
        result = utils.eliminate_val(definition, target)
        self.assertIsNone(result)

    def test_found_the_lowest(self):
        definition = "1,2,3"
        target = [1]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("2,3", result)

    def test_found_the_highest(self):
        definition = "1,2,3"
        target = [3]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("1,2", result)

    def test_found_range_of_low(self):
        definition = "1-10"
        target = [1]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("2-10", result)

    def test_found_range_of_highest_next_to_lowest(self):
        definition = "1-2"
        target = [2]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("1", result)

    def test_found_range_of_lowest_next_to_highest(self):
        definition = "1-2"
        target = [1]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("2", result)

    def test_found_range_of_high(self):
        definition = "1-10"
        target = [10]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("1-9", result)

    def test_found_between_ranges(self):
        definition = "1-10"
        target = [5]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("1-4,6-10", result)

    def test_found_between_ranges_next_to_the_lowest(self):
        definition = "1-10"
        target = [2]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("1,3-10", result)

    def test_found_between_ranges_next_to_the_highest(self):
        definition = "1-10"
        target = [9]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("10,1-8", result)

    def test_found_between_ranges_next_to_the_lowest_and_highest(self):
        definition = "1-3"
        target = [2]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("1,3", result)

    def test_matches_only_one(self):
        definition = "100"
        target = [100]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("", result)

    def test_no_matched(self):
        definition = "10,11"
        target = [100]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("10,11", result)

    def test_no_matched_with_boundary(self):
        definition = "1-10"
        target = [100]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("1-10", result)

    def test_specify_target_with_list_and_include_1(self):
        definition = "1-5"
        target = [1, 10]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("2-5", result)

    def test_specify_target_with_list_and_include_2(self):
        definition = "1-5"
        target = [1, 2]
        result = utils.eliminate_val(definition, target)
        self.assertEqual("3-5", result)


class DummyNetwork(object):

    def __init__(self, segments):
        self.network_segments = segments


class TestGetNetworkSegments(FujitsuCommonUtilsTestCase):
    """Test class for following methods

       * get_network_type
       * get_segmentation_id
       * get_physical_network
    """

    def setUp(self):
        super(TestGetNetworkSegments, self).setUp()

        self.vlan_segments = [{
            "network_type": "vlan",
            "physical_network": 'physnet1',
            "segmentation_id": 100}]
        self.vxlan_segments = [{
            "network_type": "vxlan",
            "segmentation_id": 10000}]
        self.flat_segments = [{
            "network_type": "flat",
            "physical_network": 'flatnet'}]
        self.ctx_vlan = DummyNetwork(self.vlan_segments)
        self.ctx_vxlan = DummyNetwork(self.vxlan_segments)
        self.ctx_flat = DummyNetwork(self.flat_segments)
        self.attrs = ['NETWORK_TYPE', 'PHYSICAL_NETWORK', 'SEGMENTATION_ID']

    def test_get_from_network_dict(self):
        network = {
            "provider:network_type": "vlan",
            "provider:physical_network": 'physnet1',
            "provider:segmentation_id": 100
        }
        retvals = [
            utils._get_provider_attribute(network, attr) for attr in self.attrs
        ]
        self.assertEqual("vlan", retvals[0])
        self.assertEqual('physnet1', retvals[1])
        self.assertEqual(100, retvals[2])

    def test_get_from_network_dict_with_segments(self):
        net_dict = {
            "segments": [{
                "provider:network_type": "vlan",
                "provider:physical_network": 'physnet1',
                "provider:segmentation_id": 100
            }]
        }
        retvals = [
            utils._get_provider_attribute(
                net_dict, attr) for attr in self.attrs]
        self.assertEqual("vlan", retvals[0])
        self.assertEqual('physnet1', retvals[1])
        self.assertEqual(100, retvals[2])

    def test_get_from_network_context_with_segments(self):
        retvals = [
            utils._get_provider_attribute(
                self.ctx_vlan, attr) for attr in self.attrs]
        self.assertEqual("vlan", retvals[0])
        self.assertEqual('physnet1', retvals[1])
        self.assertEqual(100, retvals[2])

    def test_get_from_network_context_with_segments_provider(self):
        retvals = [
            utils._get_provider_attribute(
                self.ctx_vlan, attr) for attr in self.attrs]
        self.assertEqual("vlan", retvals[0])
        self.assertEqual('physnet1', retvals[1])
        self.assertEqual(100, retvals[2])

    def test_get_from_network_context_vxlan(self):
        self.assertEqual('vxlan', utils.get_network_type(self.ctx_vxlan))
        self.assertEqual(10000, utils.get_segmentation_id(self.ctx_vxlan))
        self.assertIsNone(utils.get_physical_network(self.ctx_vxlan))

    def test_get_from_network_context_flat(self):
        self.assertEqual('flat', utils.get_network_type(self.ctx_flat))
        self.assertEqual('flatnet', utils.get_physical_network(self.ctx_flat))
        self.assertIsNone(utils.get_segmentation_id(self.ctx_flat))

    def test_illegal_attrs_not_exist(self):
        net_dict = {}
        self.assertIsNone(utils.get_network_type(net_dict))
        self.assertIsNone(utils.get_physical_network(net_dict))
        self.assertIsNone(utils.get_segmentation_id(net_dict))

    def test_illegal_unable_to_load_attributes(self):
        for attr in [self.ctx_vlan, self.ctx_vxlan, self.ctx_flat]:
            self.assertRaises(
                AttributeError,
                utils._get_provider_attribute,
                attr,
                'non_exist_attribute')


class TestGetPhysicalConnectivity(FujitsuCommonUtilsTestCase):
    """Test class for get_physical_connectivity"""

    def test_get_correct_data(self):
        local_link_info = utils.get_physical_connectivity(self.port)
        self.assertEqual(self.lli, local_link_info)

    def test_local_link_info_is_undefined(self):
        del self.port['binding:profile']['local_link_information']
        local_link_info = utils.get_physical_connectivity(self.port)
        self.assertEqual([], local_link_info)

    def test_local_link_info_is_empty(self):
        self.port['binding:profile']['local_link_information'] = []
        local_link_info = utils.get_physical_connectivity(self.port)
        self.assertEqual([], local_link_info)

    def test_some_key_in_local_link_info_is_missing(self):
        for missing in self.lli[1]:
            tmp = copy.copy(self.lli[1])
            del tmp[missing]
            self.port['binding:profile']['local_link_information'] = [tmp]
            local_link_info = utils.get_physical_connectivity(self.port)
            self.assertEqual([], local_link_info)

    def test_some_value_in_local_link_info_is_missing(self):
        for missing in self.lli[0]:
            tmp = copy.copy(self.lli[0])
            tmp[missing] = ""
            self.port['binding:profile']['local_link_information'] = [tmp]
            local_link_info = utils.get_physical_connectivity(self.port)
            self.assertEqual([], local_link_info)


class TestIsBaremetalDeploy(FujitsuCommonUtilsTestCase):
    """Test class for is_baremetal_deploy"""

    def test_vnic_type_is_baremetal(self):
        self.assertTrue(utils.is_baremetal(self.port))

    def test_vnic_type_is_not_baremetal(self):
        self.port[portbindings.VNIC_TYPE] = portbindings.VNIC_NORMAL
        self.assertFalse(utils.is_baremetal(self.port))

        self.port[portbindings.VNIC_TYPE] = None
        self.assertFalse(utils.is_baremetal(self.port))

    def test_illegal_port_without_vnic_type(self):
        self.assertFalse(utils.is_baremetal({}))

    def test_illegal_argument_is_none(self):
        self.assertRaises(AttributeError, utils.is_baremetal, None)

    def test_illegal_argument_is_empty_list(self):
        self.assertRaises(AttributeError, utils.is_baremetal, [])


class TestIsLag(FujitsuCommonUtilsTestCase):
    """Test class for is_lag"""

    def test_lli_is_one(self):
        self.assertFalse(utils.is_lag([self.lli[0]]))

    def test_lli_is_more_than_2(self):
        self.assertTrue(utils.is_lag(self.lli))

    def test_lli_is_empty(self):
        self.assertFalse(utils.is_lag([]))



#!/usr/bin/env python

# Copyright 2014 Hewlett-Packard Development Company, L.P.
# Copyright 2014 Samsung Electronics
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Trace a subunit stream in reasonable detail and high accuracy."""

import argparse
import functools
import os
import re
import sys

import mimeparse
import subunit
import testtools

DAY_SECONDS = 60 * 60 * 24
FAILS = []
RESULTS = {}


class Starts(testtools.StreamResult):

    def __init__(self, output):
        super(Starts, self).__init__()
        self._output = output

    def startTestRun(self):
        self._neednewline = False
        self._emitted = set()

    def status(self, test_id=None, test_status=None, test_tags=None,
               runnable=True, file_name=None, file_bytes=None, eof=False,
               mime_type=None, route_code=None, timestamp=None):
        super(Starts, self).status(
            test_id, test_status,
            test_tags=test_tags, runnable=runnable, file_name=file_name,
            file_bytes=file_bytes, eof=eof, mime_type=mime_type,
            route_code=route_code, timestamp=timestamp)
        if not test_id:
            if not file_bytes:
                return
            if not mime_type or mime_type == 'test/plain;charset=utf8':
                mime_type = 'text/plain; charset=utf-8'
            primary, sub, parameters = mimeparse.parse_mime_type(mime_type)
            content_type = testtools.content_type.ContentType(
                primary, sub, parameters)
            content = testtools.content.Content(
                content_type, lambda: [file_bytes])
            text = content.as_text()
            if text and text[-1] not in '\r\n':
                self._neednewline = True
            self._output.write(text)
        elif test_status == 'inprogress' and test_id not in self._emitted:
            if self._neednewline:
                self._neednewline = False
                self._output.write('\n')
            worker = ''
            for tag in test_tags or ():
                if tag.startswith('worker-'):
                    worker = '(' + tag[7:] + ') '
            if timestamp:
                timestr = timestamp.isoformat()
            else:
                timestr = ''
                self._output.write('%s: %s%s [start]\n' %
                                   (timestr, worker, test_id))
            self._emitted.add(test_id)


def cleanup_test_name(name, strip_tags=True, strip_scenarios=False):
    """Clean up the test name for display.

    By default we strip out the tags in the test because they don't help us
    in identifying the test that is run to it's result.

    Make it possible to strip out the testscenarios information (not to
    be confused with tempest scenarios) however that's often needed to
    indentify generated negative tests.
    """
    if strip_tags:
        tags_start = name.find('[')
        tags_end = name.find(']')
        if tags_start > 0 and tags_end > tags_start:
            newname = name[:tags_start]
            newname += name[tags_end + 1:]
            name = newname

    if strip_scenarios:
        tags_start = name.find('(')
        tags_end = name.find(')')
        if tags_start > 0 and tags_end > tags_start:
            newname = name[:tags_start]
            newname += name[tags_end + 1:]
            name = newname

    return name


def get_duration(timestamps):
    start, end = timestamps
    if not start or not end:
        duration = ''
    else:
        delta = end - start
        duration = '%d.%06ds' % (
            delta.days * DAY_SECONDS + delta.seconds, delta.microseconds)
    return duration


def find_worker(test):
    for tag in test['tags']:
        if tag.startswith('worker-'):
            return int(tag[7:])
    return 'NaN'


# Print out stdout/stderr if it exists, always
def print_attachments(stream, test, all_channels=False):
    """Print out subunit attachments.

    Print out subunit attachments that contain content. This
    runs in 2 modes, one for successes where we print out just stdout
    and stderr, and an override that dumps all the attachments.
    """
    channels = ('stdout', 'stderr')
    for name, detail in test['details'].items():
        # NOTE(sdague): the subunit names are a little crazy, and actually
        # are in the form pythonlogging:'' (with the colon and quotes)
        name = name.split(':')[0]
        if detail.content_type.type == 'test':
            detail.content_type.type = 'text'
        if (all_channels or name in channels) and detail.as_text():
            title = "Captured %s:" % name
            stream.write("\n%s\n%s\n" % (title, ('~' * len(title))))
            # indent attachment lines 4 spaces to make them visually
            # offset
            for line in detail.as_text().split('\n'):
                stream.write("    %s\n" % line)


def show_outcome(stream, test, print_failures=False, failonly=False):
    global RESULTS
    status = test['status']
    # TODO(sdague): ask lifeless why on this?
    if status == 'exists':
        return

    worker = find_worker(test)
    name = cleanup_test_name(test['id'])
    duration = get_duration(test['timestamps'])

    if worker not in RESULTS:
        RESULTS[worker] = []
    RESULTS[worker].append(test)

    # don't count the end of the return code as a fail
    if name == 'process-returncode':
        return

    if status == 'fail':
        FAILS.append(test)
        stream.write('{%s} %s [%s] ... FAILED\n' % (
            worker, name, duration))
        if not print_failures:
            print_attachments(stream, test, all_channels=True)
    elif not failonly:
        if status == 'success':
            stream.write('{%s} %s [%s] ... ok\n' % (
                worker, name, duration))
            print_attachments(stream, test)
        elif status == 'skip':
            stream.write('{%s} %s ... SKIPPED: %s\n' % (
                worker, name, test['details']['reason'].as_text()))
        else:
            stream.write('{%s} %s [%s] ... %s\n' % (
                worker, name, duration, test['status']))
            if not print_failures:
                print_attachments(stream, test, all_channels=True)

    stream.flush()


def print_fails(stream):
    """Print summary failure report.

    Currently unused, however there remains debate on inline vs. at end
    reporting, so leave the utility function for later use.
    """
    if not FAILS:
        return
    stream.write("\n==============================\n")
    stream.write("Failed %s tests - output below:" % len(FAILS))
    stream.write("\n==============================\n")
    for f in FAILS:
        stream.write("\n%s\n" % f['id'])
        stream.write("%s\n" % ('-' * len(f['id'])))
        print_attachments(stream, f, all_channels=True)
    stream.write('\n')


def count_tests(key, value):
    count = 0
    for k, v in RESULTS.items():
        for item in v:
            if key in item:
                if re.search(value, item[key]):
                    count += 1
    return count


def run_time():
    runtime = 0.0
    for k, v in RESULTS.items():
        for test in v:
            runtime += float(get_duration(test['timestamps']).strip('s'))
    return runtime


def worker_stats(worker):
    tests = RESULTS[worker]
    num_tests = len(tests)
    delta = tests[-1]['timestamps'][1] - tests[0]['timestamps'][0]
    return num_tests, delta


def print_summary(stream):
    stream.write("\n======\nTotals\n======\n")
    stream.write("Run: %s in %s sec.\n" % (count_tests('status', '.*'),
                                           run_time()))
    stream.write(" - Passed: %s\n" % count_tests('status', 'success'))
    stream.write(" - Skipped: %s\n" % count_tests('status', 'skip'))
    stream.write(" - Failed: %s\n" % count_tests('status', 'fail'))

    # we could have no results, especially as we filter out the process-codes
    if RESULTS:
        stream.write("\n==============\nWorker Balance\n==============\n")

        for w in range(max(RESULTS.keys()) + 1):
            if w not in RESULTS:
                stream.write(
                    " - WARNING: missing Worker %s! "
                    "Race in testr accounting.\n" % w)
            else:
                num, time = worker_stats(w)
                stream.write(" - Worker %s (%s tests) => %ss\n" %
                             (w, num, time))


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--no-failure-debug', '-n', action='store_true',
                        dest='print_failures', help='Disable printing failure '
                        'debug information in realtime')
    parser.add_argument('--fails', '-f', action='store_true',
                        dest='post_fails', help='Print failure debug '
                        'information after the stream is proccesed')
    parser.add_argument('--failonly', action='store_true',
                        dest='failonly', help="Don't print success items",
                        default=(
                            os.environ.get('TRACE_FAILONLY', False)
                            is not False))
    return parser.parse_args()


def main():
    args = parse_args()
    stream = subunit.ByteStreamToStreamResult(
        sys.stdin, non_subunit_name='stdout')
    starts = Starts(sys.stdout)
    outcomes = testtools.StreamToDict(
        functools.partial(show_outcome, sys.stdout,
                          print_failures=args.print_failures,
                          failonly=args.failonly
                          ))
    summary = testtools.StreamSummary()
    result = testtools.CopyStreamResult([starts, outcomes, summary])
    result.startTestRun()
    try:
        stream.run(result)
    finally:
        result.stopTestRun()
    if count_tests('status', '.*') == 0:
        print("The test run didn't actually run any tests")
        return 1
    if args.post_fails:
        print_fails(sys.stdout)
    print_summary(sys.stdout)
    return (0 if summary.wasSuccessful() else 1)


if __name__ == '__main__':
    sys.exit(main())
